/* cmod - set up the user environment with module files.
   Copyright (C) 1997 by the authors of cmod (see AUTHORS).

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "env.h"
#include "cmod_main.h"
#include "cmod_trace.h"
#include "cmod_parse.h"
#include "cmod_alloc.h"
#include "unused.h"

/* Forward declarations. */
static void parse_module(FILE *fp,
			 const char *filename,
			 int for_removal,
			 int *lineno);

/* Random globals */

const char *argv0 = NULL;
int cmod_errors = 0;
struct envarray *global_env = NULL;
static int warning_found = 0;




static void
warning(const char *filename,
	const int *lineno,
	const char *msg)
{
  warning_found = 1;
  if (filename != NULL && lineno != NULL)
    fprintf(stderr, "%s:%d: warning: %s\n", filename, *lineno, msg);
  else
    fprintf(stderr, "module: warning: %s\n", msg);
}


static int
append_path(const char *UNUSED(module),
	    const int *UNUSED(line),
	    int argc,
	    const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;

  assert(argc == 2);
  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_append_frags(e, ':', frags, 1);
  return 0;
}


static int
prepend_path(const char *UNUSED(module),
	     const int *UNUSED(line),
	     int argc,
	     const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;

  assert(argc == 2);
  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_prepend_frags(e, ':', frags, 1);
  return 0;
}


static int
remove_path(const char *UNUSED(module),
	    const int *UNUSED(line),
	    int argc,
	    const char *const *argv)
{
  const char *name;
  const char *frags;
  struct env *e;
  
  assert(argc == 2);

  name = argv[0];
  frags = argv[1];

  e = envarray_locate_create(global_env, name, 1);
  env_remove_frags(e, ':', frags);
  return 0;
}


static int
set_env(const char *UNUSED(module),
	const int *UNUSED(line),
	int argc,
	const char *const *argv)
{
  const char *name;
  char *val;
  struct env *e;

  assert(argc == 2);

  name = argv[0];
  val = xstrdup(argv[1]);

  e = envarray_locate_create(global_env, name, 0);
  env_setenv(e, val);
  return 0;
}


static int
unset_env(const char *UNUSED(module),
	  const int *UNUSED(line),
	  int argc,
	  const char *const *argv)
{
  const char *name;
  struct env *e;

  assert(argc == 2 || argc == 1);

  name = argv[0];

  e = envarray_locate_create(global_env, name, 0);
  env_unsetenv(e);
  return 0;
}


static void
do_use_path(const char *path)
{
  struct env *e = envarray_locate_create(global_env, "MODULEPATH", 1);
  env_prepend_frags(e, ':', path, 1);
}

static int
use_path(const char *UNUSED(containing_filename),
	 const int *UNUSED(containing_lineno),
	 int argc,
	 const char *const *argv)
{
  assert(argc == 1);
  do_use_path(argv[0]);
  return 0;
}


static void
do_unuse_path(const char *path)
{
  struct env *e = envarray_locate_create(global_env, "MODULEPATH", 1);
  env_remove_frags(e, ':', path);
}

static int
unuse_path(const char *UNUSED(containing_filename),
	 const int *UNUSED(containing_lineno),
	 int argc,
	 const char *const *argv)
{
  assert(argc == 1);
  do_unuse_path(argv[0]);
  return 0;
}


/* Open MODULE for reading and return a FILE*.  Search for MODULE in
   the environment variable MODULEPATH (which is stored in the global
   environment if it wasn't already present there).  The file name is
   returned in FOUND_FILENAME.  Ownership of the filename passes to
   the caller. */

static FILE *
open_module(const char *module,
	    char **found_filename)
{
  int component = 0;
  const char *dir;
  FILE *fp;
  struct env *module_path;
  char *fn;

  module_path = envarray_locate_create(global_env, "MODULEPATH", 1);

  while ((dir = env_iterate(module_path, ':', &component)) != NULL)
    {
      fn = xmalloc(strlen(dir) + strlen(module) + 2);
      sprintf(fn, "%s/%s", dir, module);
      fp = fopen(fn, "r");
      if (fp != NULL)
	{
	  *found_filename = fn;
	  return fp;
	}
      xfree(fn);
    }

  fprintf(stderr, "%s: failed to find %s in $MODULEPATH.\n", argv0, module);
  exit(CMOD_EXIT_ERROR);
}

  
static void
do_rm_module(const char *containing_filename,
	     const int *containing_lineno,
	     const char *module_name)
{
  char *fn;
  int lineno = 0;
  FILE *mfp;

  if (env_frag_present(envarray_locate_create(global_env, "LOADEDMODULES", 1),
		       ':', module_name))
    {
      mfp = open_module(module_name, &fn);
      assert(mfp != NULL && fn != NULL);

      parse_module(mfp, fn, 1, &lineno);
      env_remove_one_frag(envarray_locate_create(global_env, "LOADEDMODULES",
						 1),
			  ':', module_name);
      fclose(mfp);
      xfree(fn);
    }
  else
    warning(containing_filename, containing_lineno,
	    "ignored attempt to unload unloaded module.");
}

static int
rm_module(const char *containing_filename,
	  const int *containing_lineno,
	  int argc,
	  const char *const *argv)
{
  int opti;

  for (opti = 0; opti < argc; ++opti)
    do_rm_module(containing_filename, containing_lineno, argv[opti]);
  return 0;
}


static void
do_add_module(const char *containing_filename,
	      const int *containing_lineno,
	      const char *module_name)
{
  int lineno = 0;
  FILE *mfp;
  char *fn;
  struct env *loaded_modules;
  
  if (env_frag_present(envarray_locate_create(global_env, "LOADEDMODULES", 1),
		       ':', module_name))
    do_rm_module(containing_filename, containing_lineno, module_name);

  mfp = open_module(module_name, &fn);
  assert(mfp != NULL && fn != NULL);

  parse_module(mfp, fn, 0, &lineno);

  loaded_modules = envarray_locate_create(global_env, "LOADEDMODULES", 1);
  env_append_one_frag(loaded_modules, ':', module_name);

  fclose(mfp);
  xfree(fn);
}


static int
add_module(const char *containing_filename,
	   const int *containing_lineno,
	   int argc,
	   const char *const *argv)
{
  int opti;

  for (opti = 0; opti < argc; ++opti)
    do_add_module(containing_filename, containing_lineno, argv[opti]);
  return 0;
}



static int
list_modules(const char *UNUSED(from_module),
	     const int *UNUSED(from_line),
	     int argc,
	     const char *const *UNUSED(argv))
{
  struct env *loaded_modules;
  int i;
  const char *module;

  assert(argc == 0);

  loaded_modules = envarray_locate_create(global_env, "LOADEDMODULES", 1);

  fprintf(stderr,"Currently loaded modules:\n");
  for (i = 0; (module = env_iterate(loaded_modules, ':', &i)) != NULL; )
    fprintf(stderr,"%3d) %s\n", i, module);

  return 0;
}

static int
no_op(const char *UNUSED(module),
      const int *UNUSED(line),
      int UNUSED(argc),
      const char *const *UNUSED(argv))
{
  return 0;
}

static const struct cmd cmd_table[] = {
  {"append-path", NULL, 2, 2, append_path, remove_path},
  {"prepend-path", NULL, 2, 2, prepend_path, remove_path},
  {"remove-path", NULL, 2, 2, remove_path, remove_path},
  {"setenv", NULL, 2, 2, set_env, unset_env},
  {"unsetenv", NULL, 1, 1, unset_env, unset_env},
  {"module", "add", 1, -1, add_module, rm_module},
  {"module", "load", 1, -1, add_module, rm_module},
  {"module", "unload", 1, -1, rm_module, add_module},
  {"module", "use", 1, 1, use_path, unuse_path},
  {"module", "unuse", 1, 1, unuse_path, use_path},
  {"module", "list", 0, 0, list_modules, no_op},
  {NULL, NULL, 0, 0, NULL, NULL}
};

static void
parse_module(FILE *fp,
	     const char *filename,
	     int for_removal,
	     int *lineno)
{
  /* FIXME: For now, discard the error status. */
  cmod_parse_module(fp, filename, for_removal, cmd_table, lineno);
}


static void
usage(void)
{
  fprintf(stderr,
	  "Usage: module sub-command [args...]\n\n");
  fprintf(stderr,
	  "  Available sub-commands:\n\n");
  fprintf(stderr,
	  "    add|load  modulefile [modulefile...]    Add module\n");
  fprintf(stderr,
	  "    rm|unload modulefile [modulefile...]    Remove module\n");
  fprintf(stderr,
	  "    use       dir                           Add to MODULEPATH\n");
  fprintf(stderr,
	  "    unuse     dir                           Remove from MODULEPATH"
	  "\n");
  fprintf(stderr,
	  "    list                                    List loaded modules\n");
  fprintf(stderr,"\n");
  exit(CMOD_EXIT_ERROR);
}


int
main(int argc,
     const char **argv)
{
  argv0 = argv[0];

#if 0
  /* FIXME: Add a switch to enable selected traces. */
  set_traces(TRACE_ASSIGN);
#endif

  if (argc < 2)
    usage();

  if (!strcmp(argv[1], "sh") || !strcmp(argv[1], "csh"))
    {
    }
  else
    usage();

  global_env = create_envarray();

  switch (cmod_parse_argv(cmd_table, argc-2, argv+2))
    {
    case -1:
      /* bad usage */
      usage();
    case 0:
      /* no error */
      break;
    default:
      /* error (but command line looks ok) */
      break;
    }

  if (cmod_errors > 0)
    exit(CMOD_EXIT_ERROR);

  envarray_dump(global_env, stdout, argv[1]);
  destroy_envarray(global_env);

  if (warning_found)
    exit(CMOD_EXIT_WARNING);
  else
    exit(0);
}
